#define _CRT_SECURE_NO_WARNINGS
#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#define _USE_MATH_DEFINES
#include <cmath>
#include <vector>
#include "include.h"
#include <time.h>


#define MIN(a,b)             (((a) < (b)) ? (a) : (b))
#define MAX(a,b)             (((a) > (b)) ? (a) : (b))
#define CLAMP(a, mi,ma)      MIN(MAX(a,mi),ma)

char *vram = NULL;

typedef struct
{
	float x, y;
} vec2;

typedef struct
{
	float x, y, z;
} vec3;



// Utility functions. Because C doesn't have templates,
// we use the slightly less safe preprocessor macros to
// implement these functions that work with multiple types.
#define VXS(x0,y0, x1,y1)    ((x0)*(y1) - (x1)*(y0))   // vxs: 2d Vector cross product



// PointSide: Determine which side of a line the point is on. Return value: <0, =0 or >0.
float PointSide(float px, float py, float x0, float y0, float x1, float y1)
{
	return VXS((x1)-(x0), (y1)-(y0), (px)-(x0), (py)-(y0));
}

// Overlap:  Determine whether the two number ranges overlap.
bool Overlap(float a0, float a1, float b0, float b1)
{
	return (MIN(a0, a1) <= MAX(b0, b1) && MIN(b0, b1) <= MAX(a0, a1));
}

// IntersectBox: Determine whether two 2D-boxes intersect.
bool IntersectBox(float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3)
{
	return (Overlap(x0, x1, x2, x3) && Overlap(y0, y1, y2, y3));
}


// Intersect: Calculate the point of intersection between two lines.
void Intersect(float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4, vec2 *rval)
{
	float xi = VXS(VXS(x1, y1, x2, y2), (x1)-(x2), VXS(x3, y3, x4, y4), (x3)-(x4)) / VXS((x1)-(x2), (y1)-(y2), (x3)-(x4), (y3)-(y4));
	float yi = VXS(VXS(x1, y1, x2, y2), (y1)-(y2), VXS(x3, y3, x4, y4), (y3)-(y4)) / VXS((x1)-(x2), (y1)-(y2), (x3)-(x4), (y3)-(y4));

	rval->x = x1;
	rval->y = x1;
}


// Player: location
struct player
{
	vec3 position;      // Current position
	vec3 velocity;   // Current motion vector
	float angle, anglesin, anglecos, yaw;   // Looking towards (and sin() and cos() thereof)
	unsigned sector;                        // Which sector the player is currently in
};

struct player player;



unsigned int cxClient;
unsigned int cyClient;
#define WM_TICK 1

void draw_pixels(HDC hdc, HDC hdcMem, int width, int height, waddata_t &data);
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

char *get_file(char *filename, unsigned int *size)
{
	FILE	*file;
	char	*buffer;
	int	file_size, bytes_read;

	file = fopen(filename, "rb");
	if (file == NULL)
		return 0;
	fseek(file, 0, SEEK_END);
	file_size = ftell(file);
	fseek(file, 0, SEEK_SET);
	buffer = new char[file_size + 1];
	bytes_read = (int)fread(buffer, sizeof(char), file_size, file);
	if (bytes_read != file_size)
	{
		delete[] buffer;
		fclose(file);
		return 0;
	}
	fclose(file);
	buffer[file_size] = '\0';

	if (size != NULL)
	{
		*size = file_size;
	}

	return buffer;
}


#define MAX_LEVEL 9
char level_name[MAX_LEVEL][80] = {
	"E1M1",
	"E1M2",
	"E1M3",
	"E1M4",
	"E1M5",
	"E1M6",
	"E1M7",
	"E1M8",
	"E1M9",
};


void load_wad(waddata_t *data, int level)
{
	int lump_size;
	char *pdata = NULL;

	data->pvert = (vert_t *)get_wadfile("DOOM1.WAD", level_name[level], "VERTEXES", &lump_size, &pdata);
	data->num_vert = lump_size / (sizeof(vert_t));

	data->pline = (linedef_t *)get_wadfile("DOOM1.WAD", level_name[level], "LINEDEFS", &lump_size, &pdata);
	data->num_line = lump_size / (sizeof(linedef_t));

	data->pthing = (thing_t *)get_wadfile("DOOM1.WAD", level_name[level], "THINGS", &lump_size, &pdata);
	data->num_thing = lump_size / (sizeof(thing_t));

	data->pside = (sidedef_t *)get_wadfile("DOOM1.WAD", level_name[level], "SIDEDEFS", &lump_size, &pdata);
	data->num_side = lump_size / (sizeof(sidedef_t));

	data->psector = (sector_t *)get_wadfile("DOOM1.WAD", level_name[level], "SECTORS", &lump_size, &pdata);
	data->num_sector = lump_size / (sizeof(sector_t));

	data->pseg = (seg_t *)get_wadfile("DOOM1.WAD", level_name[level], "SEGS", &lump_size, &pdata);
	data->num_seg = lump_size / (sizeof(seg_t));

	data->psubsector = (subsector_t *)get_wadfile("DOOM1.WAD", level_name[level], "SSECTORS", &lump_size, &pdata);
	data->num_subsector = lump_size / (sizeof(subsector_t));

	data->pnode = (node_t *)get_wadfile("DOOM1.WAD", level_name[level], "NODES", &lump_size, &pdata);
	data->num_node = lump_size / (sizeof(node_t));
}

typedef struct
{
	float x;
	float y;
	float scale;
} offset_t;

offset_t level_offsets[MAX_LEVEL] = {
	{ 1180, 4880, 2.5f },
	{ 4700, 1460, 1.5f },
	{ 4720, 3740, 2.0f },
	{ 2680,  640, 2.5 },
	{ 3780, 920, 2.0f },
	{ 4780, 2420, 1.5 },
	{ 4140, 2840, 2.0f },
	{ 6360, 520, 1.0f },
	{ 3220, 1600, 2.0 }
};


char info[512];
int draw_count = 0;

void DrawSubsector(waddata_t &data, short index)
{
	int j;
	float scale = data.scale;
	float offset_x = data.offset_x;
	float offset_y = data.offset_y;

	if (index < 0 || index >= data.num_subsector)
	{
		printf("Fail\r\n");
		return;
	}


	if (draw_count >= data.max_draw)
	{
		return;
	}
	draw_count++;

	int first = data.psubsector[index].first_seg;
	int count = data.psubsector[index].num_seg;
	SelectObject(data.hdc, data.blue);

	for (j = first; j < first + count; j++)
	{
		int a = data.pseg[j].start_vert;
		int b = data.pseg[j].end_vert;

		float startx = (data.pvert[a].x * scale + offset_x * scale) / 10.0f;
		float starty = (data.pvert[a].y * scale + offset_y * scale) / 10.0f;
		float endx = (data.pvert[b].x * scale + offset_x * scale) / 10.0f;
		float endy = (data.pvert[b].y * scale + offset_y * scale) / 10.0f;

		MoveToEx(data.hdc, startx, starty, NULL);
		LineTo(data.hdc, endx, endy);
	}

}

int PointOnSide(float x, float y, node_t *node)
{

#if 0
	if (PointSide(x, y, node->x, node->y, node->x + node->dx, node->y + node->dy) > 0)
		return 1;
	else
		return 0;
#else
	float dx;
	float dy;
	float left;
	float right;

	if (node->dx == 0)
	{
	if (x <= node->x)
	return node->dy > 0;

	return node->dy < 0;
	}
	if (node->dy == 0)
	{
	if (y <= node->y)
	return node->dx < 0;

	return node->dx > 0;
	}

	dx = (x - node->x);
	dy = (y - node->y);

	left = node->dy * dx;
	right = dy * node->dx;

	if (right < left)
	{
	// front side
	return 0;
	}
	// back side
	return 1;
#endif
}

bool CheckBBox(short *bbox)
{
	short x0, y0, x1, y1;


	// top, bottom, left, right
	y0 = bbox[0];
	y1 = bbox[1];
	x0 = bbox[2];
	x1 = bbox[3];
	//bool IntersectBox(float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3)
	return IntersectBox(x0, y0, x1, y1, (short)player.position.x, (short)player.position.y, (short)player.position.x + 1, (short)player.position.y + 1);
}


void RenderBSPNode(waddata_t &data, short bspnum, int i)
{
	node_t *node;
	int		side;

	// Found a leaf (subsector), draw it
	if (bspnum & 0x8000)
	{
		if (bspnum == -1)
			DrawSubsector(data, 0);
		else
			DrawSubsector(data, bspnum & 0x7FFF);
		return;
	}

	node = &data.pnode[bspnum];

	// Decide which side the view point is on.
	side = PointOnSide(player.position.x, player.position.y, node);
	i++;

	// Transform World->Client
	float startx = (node->x * data.scale + data.offset_x * data.scale) / 10.0f;
	float starty = (node->y * data.scale + data.offset_y * data.scale) / 10.0f;
	float endx = ((node->x + node->dx) * data.scale + data.offset_x * data.scale) / 10.0f;
	float endy = ((node->y + node->dy) * data.scale + data.offset_y * data.scale) / 10.0f;

	if (side == 0)
	{
		SelectObject(data.hdc, data.green[i * 5]);
	}
	else
	{
		SelectObject(data.hdc, data.green_dash[i * 5]);
	}

	char pnum[80];
	sprintf(pnum, "P%d", i);
	RECT rect;

	rect.left = startx;
	rect.top = starty - 25;
	rect.right = startx + 200;
	rect.bottom = starty;

	// Draw BSP lines
	DrawText(data.hdc, pnum, -1, &rect, 0);
	MoveToEx(data.hdc, startx, starty, NULL);
	LineTo(data.hdc, endx, endy);


	SelectObject(data.hdc, data.cross_brush);
	// mask X axial bsp planes
	if (GetAsyncKeyState(VK_SPACE))
	{
		if (startx - endx == 0)
		{
			if (startx <= endx && starty <= endy)
			{
				if (side == 0)
					Rectangle(data.hdc, 0, 0, startx, cyClient);
				else
					Rectangle(data.hdc, startx, 0, cxClient, cyClient);
			}
			else if (startx <= endx && starty > endy)
			{
				if (side == 1)
					Rectangle(data.hdc, 0, 0, startx, cyClient);
				else
					Rectangle(data.hdc, startx, 0, cxClient, cyClient);
			}

		}

		// mask Y axial bsp planes
		if (starty - endy == 0)
		{
			if (starty <= endy && startx <= endx)
			{
				if (side == 1)
					Rectangle(data.hdc, 0, 0, cxClient, starty);
				else
					Rectangle(data.hdc, 0, starty, cxClient, cyClient);
			}
			else if (starty <= endy && startx > endx)
			{
				if (side == 0)
					Rectangle(data.hdc, 0, 0, cxClient, starty);
				else
					Rectangle(data.hdc, 0, starty, cxClient, cyClient);
			}

		}
	}

	SelectObject(data.hdc, WHITE_BRUSH);


	if (i >= data.max_depth)
		return;


	// Recursively divide front space.
	RenderBSPNode(data, node->child[side], i);

	// Possibly divide back space.
	if (CheckBBox(node->bbox[side ^ 1]) || data.draw_order == true)
	{
		RenderBSPNode(data, node->child[side ^ 1], i);
	}
}


void Draw(HDC hdc, waddata_t &data)
{
	int i;
	SelectObject(hdc, WHITE_BRUSH);
	Rectangle(hdc, 0, 0, cxClient, cyClient);
	RECT rect;




	data.hdc = hdc;
	rect.left = 0;
	rect.top = 0;
	rect.bottom = 50;
	rect.right = cxClient;
	sprintf(info, "MaxDepth %d MaxDraw %d Player (%d,%d)", data.max_depth, data.max_draw, (int)player.position.x, (int)player.position.y);


	//		sprintf(info, "%s scale %f offsetx %f offsety %f", level_name[level], scale, offset_x, offset_y);
	DrawText(hdc, info, -1, &rect, 0);



	// draw vertices
	SelectObject(hdc, GetStockObject(BLACK_PEN));
	for (i = 0; i < data.num_vert; i += 2)
	{
		// Transform World->Client
		float x = (data.pvert[i].x * data.scale + data.offset_x * data.scale) / 10.0f;
		float y = (data.pvert[i].y * data.scale + data.offset_y * data.scale) / 10.0f;

		Rectangle(hdc,
			(x - 3.0f),
			(y - 3.0f),
			(x + 3.0f),
			(y + 3.0f));
	}

	// draw lines
	for (i = 0; i < data.num_line; i++)
	{
		int a = data.pline[i].start_vert;
		int b = data.pline[i].end_vert;

		float startx = (data.pvert[a].x * data.scale + data.offset_x * data.scale) / 10.0f;
		float starty = (data.pvert[a].y * data.scale + data.offset_y * data.scale) / 10.0f;
		float endx = (data.pvert[b].x * data.scale + data.offset_x * data.scale) / 10.0f;
		float endy = (data.pvert[b].y * data.scale + data.offset_y * data.scale) / 10.0f;

		MoveToEx(hdc, startx, starty, NULL);
		LineTo(hdc, endx, endy);
	}

	// draw things
	SelectObject(hdc, data.red);
	for (i = 0; i < data.num_thing; i++)
	{
		float x = (data.pthing[i].x * data.scale + data.offset_x * data.scale) / 10.0f;
		float y = (data.pthing[i].y * data.scale + data.offset_y * data.scale) / 10.0f;


		Rectangle(hdc,
			(x - 3.0f),
			(y - 3.0f),
			(x + 3.0f),
			(y + 3.0f));
	}


	draw_count = 0;
	SelectObject(hdc, data.blue);
	RenderBSPNode(data, data.num_node - 1, 0);



	// draw player
	{
		float x = (player.position.x * data.scale + data.offset_x * data.scale) / 10.0f;
		float y = (player.position.y * data.scale + data.offset_y * data.scale) / 10.0f;


		Rectangle(hdc,
			(x - 3.0f),
			(y - 3.0f),
			(x + 3.0f),
			(y + 3.0f));
	}
}

void draw_pixels(HDC hdc, HDC hdcMem, int width, int height, waddata_t &data)
{
	HBITMAP hBitmap, hOldBitmap;

	hBitmap = CreateCompatibleBitmap(hdc, width, height);
	SetBitmapBits(hBitmap, sizeof(int) * width * height, vram);
	hdcMem = CreateCompatibleDC(hdc);
	hOldBitmap = (HBITMAP)SelectObject(hdcMem, hBitmap);
	Draw(hdcMem, data);
	// This scaling is a little strange because Stretch maintains aspect ratios
	StretchBlt(hdc, 0, 0, cxClient, cyClient, hdcMem, 0, 0, width, height, SRCCOPY);
	SelectObject(hdcMem, hOldBitmap);
	DeleteDC(hdcMem);
	DeleteObject(hBitmap);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static RECT	rect = { 0, 0, 1, 1 };
	static HDC	hdcMem;
	static int	old_style = 0;
	static int	new_style = WS_CHILD | WS_VISIBLE;
	static int	xres, yres;
	HDC		hdc;
	PAINTSTRUCT	ps;
	static bool show_cursor = true;
	static POINT center;
	static int level = 0;
	static waddata_t data;

	switch (message)
	{
	case WM_CREATE:
	{
		load_wad(&data, level);

		data.scale = level_offsets[level].scale;
		data.offset_x = level_offsets[level].x;
		data.offset_y = level_offsets[level].y;
		data.max_depth = 100;
		data.max_draw = 250;
		data.draw_order = true;

		data.red = CreatePen(PS_SOLID, 0, RGB(255, 0, 0));
		data.blue = CreatePen(PS_SOLID, 3, RGB(0, 0, 255));
		data.cross_brush = CreateHatchBrush(HS_CROSS, RGB(64, 64, 64));

		srand(time(0));
		for (int i = 0; i < 255; i++)
		{
			data.green[i] = CreatePen(PS_SOLID, 2, RGB(rand() % 255 - i, rand() % 255 - i, rand() % 255 - i));
		}
		for (int i = 0; i < 255; i++)
		{
			data.green_dash[i] = CreatePen(PS_DOT, 1, RGB(rand() % 255 - 1, rand() % 255 - i, rand() % 255 - i));
		}



		HMONITOR hmon;
		MONITORINFO mi = { sizeof(MONITORINFO) };

		hmon = MonitorFromWindow(hwnd, MONITOR_DEFAULTTONEAREST);
		GetMonitorInfo(hmon, &mi);

		xres = cxClient;
		yres = cyClient;

		vram = (char *)malloc(xres * yres * 4);
		SetTimer(hwnd, WM_TICK, 16, NULL);

		//_beginthread(start, 0, NULL);

		return 0;
	}
	case WM_MOUSEMOVE:
	{
		int lx = LOWORD(lParam);
		int ly = HIWORD(lParam);

		// Transform Client->World
		player.position.x = lx * 10.0f / data.scale - data.offset_x;
		player.position.y = ly * 10.0f / data.scale - data.offset_y;
	}
	return 0;
	case WM_TIMER:
	{
		int width = cxClient;
		int height = cyClient;

		srand(time(0));
		for (int i = 0; i < 255; i++)
		{
			DeleteObject(data.green[i]);
			data.green[i] = CreatePen(PS_SOLID, 2, RGB(rand() % 255 - i, rand() % 255 - i, rand() % 255 - i));
		}
		for (int i = 0; i < 255; i++)
		{
			DeleteObject(data.green_dash[i]);
			data.green_dash[i] = CreatePen(PS_DOT, 1, RGB(rand() % 255 - 1, rand() % 255 - i, rand() % 255 - i));
		}
		InvalidateRect(hwnd, 0, TRUE);
		return 0;
	}
	case WM_ERASEBKGND:
		return TRUE;
	case WM_SIZE:
		cxClient = LOWORD(lParam);
		cyClient = HIWORD(lParam);
		rect.right = cxClient;
		rect.bottom = cyClient;

		center.x = cxClient / 2;
		center.y = cyClient / 2;



		return 0;
	case WM_KEYDOWN:
	case WM_KEYUP:
	{
		bool pressed = (message == WM_KEYDOWN) ? true : false;


		switch (wParam)
		{
		case VK_ESCAPE:
			if (pressed)
			{
				old_style = SetWindowLong(hwnd, GWL_STYLE, new_style);
				new_style = old_style;
				SetWindowPos(hwnd, HWND_TOPMOST, 0, 0, xres, yres, 0);
			}
			break;

		case VK_NEXT:
			data.scale += 0.5f;
			break;
		case VK_PRIOR:
			data.scale -= 0.5f;
			break;
		case VK_ADD:
			level++;
			if (level >= MAX_LEVEL)
				level = MAX_LEVEL - 1;

			load_wad(&data, level);
			data.scale = level_offsets[level].scale;
			data.offset_x = level_offsets[level].x;
			data.offset_y = level_offsets[level].y;

			break;
		case VK_SUBTRACT:
			level--;
			if (level < 0)
				level = 0;

			load_wad(&data, level);
			data.scale = level_offsets[level].scale;
			data.offset_x = level_offsets[level].x;
			data.offset_y = level_offsets[level].y;
			break;

		case VK_F1:
			data.draw_order = false;
			if (data.max_depth > 0)
				data.max_depth--;
			break;
		case VK_F2:
			data.draw_order = false;
			data.max_depth++;
			break;
		case VK_F3:
			data.draw_order = true;
			data.max_depth = 250;
			if (data.max_draw > 0)
				data.max_draw--;
			break;
		case VK_F4:
			data.draw_order = true;
			data.max_depth = 250;
			data.max_draw++;
			break;

		}

		return 0;
	}
	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);
		//if (ready)
		draw_pixels(hdc, hdcMem, cxClient, cyClient, data);
		EndPaint(hwnd, &ps);
		return 0;


	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
	static TCHAR szAppName[] = TEXT("PortalLineRender");
	HWND         hwnd;
	MSG          msg;
	WNDCLASS     wndclass;

	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)(NULL_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Program requires Windows NT!"), szAppName, MB_ICONERROR);
		return 0;
	}


	hwnd = CreateWindow(szAppName, TEXT("PortalLineRender"),
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, hInstance, NULL);

	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}
